/*
**  Copyright 2006 Sun Microsystems, Inc.  All rights reserved.
**
**  mpicount.c	- Example MPShade analyzer to count dynamic instructions
**
**	This example MPShade analyzer uses the standard-tracer layer
**	functions to count the number of dynamic instructions in a set of
**	applications.  It prints the total number of instructions executed
**	by each application and also the exit code for each application.
**
**  Usage:
**
**	% mpicount -- [<application> [args] ...]
**
**	If an application is specified on the command line, it is loaded
**	and traced.  Any child processes forked by the application are
**	also traced.  If no application is specified, the analyzer waits
**	for one to attach (via the mptracer command's -attach option).
**
**	The analyzer prints (to stdout) the total number of instructions
**	executed by each application whenever the analyzer receives a
**	SIGUSR1 signal and also when all applications terminate.
*/

#pragma ident "@(#)Shade and Spixtools for Solaris 10 v1.0.1"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <shade.h>
#include <mpshade.h>

/*
 * Define a standard trace record.
 */
struct shade_trace_s {MPSHADE_TRACE};


/*
 * Define the number of instructions in each "time-slice".
 */
#define NTRSLICE	(5 * 1024)

shade_trace_t   	trbuf[8192];

unsigned long long	nInst;
volatile unsigned 	NeedPrint;

static unsigned 	parse_args(int, char **);
static void     	handle(int);
static void     	printout(void);



/*
**  Main entry point for MPShade standard-tracer analyzer.
*/
int mpshade_stdtr(
	int 	argc,
	char **	argv,
	char **	envp)
{
	int             	argc2;
	char **         	argv2;
	struct sigaction	sigact;
	unsigned        	wait_load;
	unsigned        	ntr;


	/*
	 * Parse any command line arguments.
	 */
	argc = mpshade_splitargs(argv, &argv2, &argc2);
	if (parse_args(argc, argv))
		return(1);

	/*
	 * Set up a signal handler for SIGUSR1.
	 */
	sigact.sa_handler = handle;
	sigact.sa_flags = 0;
	sigemptyset(&sigact.sa_mask);
	sigaction(SIGUSR1, &sigact, 0);

	/*
	 * If an application was specified on the command line, load it.
	 */
	if (argc2) {
		if (mpshade_load(*argv2, &argv2[1], envp) == MPSHADE_ERROR) {
			fprintf(stderr, "Unable to load application '%s'.\n",
			    *argv2);
			return(1);
		}
		wait_load = 0;
	}
	else {
		mpshade_setrunflags(MPSHADE_NOFINISH);
		wait_load = 1;
	}

	/*
	 * Trace all instructions, but log no data.
	 */
	mpshade_trsize(MPSHADE_TRSZ_TRACE);
	mpshade_trctl(mpshade_iset_newclass(SHADE_ICLASS_ANY, -1),
	    SHADE_TRI_EXECUTED, 0);
	mpshade_setmode(MPSHADE_TRACING);

	/*
	 * Set the number of trace records in each slice.  This value is not
	 * particularly important for this analyzer.
	 */
	mpshade_setslice(NTRSLICE);

	nInst = 0;
	NeedPrint = 0;

	/*
	 * If no application was specified, wait for one to attach.
	 */
	if (wait_load) {
		nInst = mpshade_run(trbuf, sizeof(trbuf));
		mpshade_setrunflags(0);
	}

	/*
	 * Trace the application(s), and count the number of executed
	 * instructions.  Print the output if we received a SIGUSR1 signal.
	 */
	while ((ntr = mpshade_run(trbuf, sizeof(trbuf))) != 0) {
		nInst += ntr;
		if (NeedPrint) {
			printout();
			NeedPrint = 0;
		}
	}

	/*
	 * All applications have terminated.  Print the output.
	 */
	printout();
	return(0);
}


/*
**  Parse any command line switches for this analyzer.
*/
static unsigned parse_args(
	int     	argc,
	char ** 	argv)
{
	/*
	 * We don't expect any switches.
	 */
	if (argc > 1) {
		fprintf(stderr, "No arguments expected\n");
		return(1);
	}
	return(0);
}


/*
**  Called if a SIGUSR1 signal is received.
*/
static void handle(
	int	sig)
{
	/*
	 * Set a flag so the output is printed next time we return to the
	 * main loop.
	 */
	NeedPrint = 1;
}


/*
**  Print the total number of instructions executed by all applications
**  and also the exit codes of any terminated applications.
*/
static void printout(void)
{
	mpshade_id_t    	idmax;
	mpshade_id_t    	id;
	mpshade_status_t	status;
	mpshade_appinfo_t	info;
	int             	i;


	printf("Total Instructions: %llu\n", nInst);

	/*
	 * Print the name of each application traced.
	 */
	idmax = mpshade_ntracers();
	for (id = 0;  id < idmax;  id++) {
		status = mpshade_appinfo(id, &info);

		printf("    [%5u] ", info.pid);
		switch (status)
		{
		case MPSHADE_STATUS_EXITED:
			printf("Exit %-3d  ", info.code);
			break;

		case MPSHADE_STATUS_SIGNALED:
			printf("Signal %-2d ", info.code);
			break;
	
		case MPSHADE_STATUS_RUNNING:
			printf("Running   ");
			break;

		case MPSHADE_STATUS_EXECED:
			printf("Untraced  ");
			break;
		}

		printf("%s", info.path);
		for (i = 0;  i < info.argc;  i++)
			printf(" %s", info.argv[i]);
		printf("\n");
	}
}
