/* Copyright 2006 Sun Microsystems, Inc.  All rights reserved. */

/*
**  brpred.c	- Shade analyzer to analyze branch prediction effects
**
**	This file defines a Shade analyzer that counts the number of
**	number of branches dynamically executed by the target application, 
**	and analyzes branch prediction efficiency.
*/

#pragma ident "@(#)Shade and Spixtools for Solaris 10 v1.0.1"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>
#include <sys/processor.h>

#include <spix_sparc.h>
#include <shade.h>
#include <shade_anal.h>
#include "getopt.h"

/* maximum number of threads in the application */
#define MAXTHDS		256
#define POWER_OF_2( N ) ( 1 << (N) )

/*
 * Define a standard trace record.
 */
struct shade_trace_s {SHADE_TRACE};

typedef struct BrpredArgs {
	/* g,n - parameters of branch prediction engine */
	int g;
	int n;
	/* trace only conditional branches */
	int condBrOnly;
	/* have we read state from file ? */
	int stateRead;
	/* simulate all threads on a single CPU */
	int singleCpuMode;  
	/* output per-cpu statistics ? */
	int perCpuStatistics;
	int rvrsGlobBrHist;		/* Store youngest branch outcomes in the MSB of the
					   global branch history */
} BrpredArgs, *BrpredArgsPtr;

/* Data structure for the (g, n) brpred branch prediction
 * scheme.  The g-bit global branch history is XORed with the
 * g bits of branch address.  This value is then used to select
 * an n-bit counter by which the prediction is made.
 */
typedef struct BrpredVars {
    
	int *theBPT;			/* The branch prediction table */
  
	int theGlobBrHist;		/* The global branch history */
  
	int brAddrMask;			/* Mask off unused branch address bits */
	int globHistMask;		/* Mask off older branch outcomes */
	int bptMinValue;		/* Lower BPT saturation value */
	int bptTakenValue;		/* BPT value above which predict taken */
	int bptMaxValue;		/* Upper BPT saturation value */
  
	long long numMispreds;		/* Number of mispredictions with brpred */
	long long numPredictions;	/* Number of brpred predicted branches */
  
} BrpredVars, *BrpredVarsPtr;

/* trace record structure with alignment enforced */
typedef union {
	double dummy;
	shade_trace_t tb[10000];
} Aligner;


/* version string */
const char	shadeuser_analversion[] = "brpred " "v1.0.1";

/* flag that signals of thread overflow */
int isThdOvfl = FALSE;

/* number of cpus being used by allpication */
int nCpusUsed = 1;

/*
 * The process ID of the first Shade process.  Trace output from all other
 * processes will be labeled as coming from a child process.
 */
static pid_t	pidFirst;

/* trace buffer */
Aligner gAligner;

/* aguments passed to brpred */
BrpredArgs gBrpredArgs;

/* prediction engines */
BrpredVars gBrpredVars[MAXTHDS];

enum{
    BRPRED_OPT_G,
    BRPRED_OPT_PCS,
    BRPRED_OPT_R,
    BRPRED_OPT_SINGLE_CPU,
    BRPRED_OPT_U
};

const shade_options_t shadeuser_opts[]={
       {"-g", "Sets the brpredspec. Usage: -g<d>,<n>. Here d is the number of bits of global branch history, used by gshare prediction scheme, and n specifies the size of counters(in bits) making up each entry in the branch prediction table", NULL, STR_PARAM},
       {"-pcs", "Provides per-cpu cache statistics for multithreaded programms", NULL, NO_PARAM},
       {"-r", "Stores youngest branch outcomes in the MSB of the global branch history.", NULL, NO_PARAM},
       {"-single-cpu", "Forces all threads to be simulated as though they were executed on a single CPU.", NULL, NO_PARAM },
       {"-u", "The outcomes of uncondition branches are not included in the global branch history.", NULL, NO_PARAM},
       {0, "", NULL, NO_PARAM}
};    


/* forward declarations of functions */
static void do_bp(shade_trace_t *tb , int ntb);
static void reset_values(BrpredVars* vars, int g, int n);
static void bind_lwps();

int
shadeuser_initialize( int argc, char **argv, char **envp ) {
  
	int i, narg, gn_read;
	int g_value;
	int n_value;

	shade_iset_t *	piset;
	shade_tset_t *	ptset;

    int                j;
    int                brpred_optnum;
    shade_option_val_t brpred_opt_value;
    
	/*
	 * Record the initial process ID.
	 */
	pidFirst = getpid();

	/* Initialize the brpred branch prediction scheme
	 * structure.
	 */
	gBrpredArgs.g = 0;
	gBrpredArgs.n = 0;
	gBrpredArgs.condBrOnly = FALSE; 
	gBrpredArgs.stateRead = FALSE;
	gBrpredArgs.rvrsGlobBrHist = FALSE; 
	gBrpredArgs.perCpuStatistics = FALSE; 
	gBrpredArgs.singleCpuMode = FALSE; 
 
	/* Parse the command-line arguments
	 */
	for ( narg = j = 1; narg < argc; narg++ ) {
    
        brpred_optnum = shade_getoptv(argc, argv, &narg, &j, shadeuser_opts, &brpred_opt_value);
        
        switch(brpred_optnum){
            case BRPRED_OPT_G:
                gn_read = sscanf( brpred_opt_value.sval, "%d,%d", &g_value, &n_value );
                break;

            case BRPRED_OPT_PCS:
                gBrpredArgs.perCpuStatistics = TRUE;
                break;

            case BRPRED_OPT_R:
                gBrpredArgs.rvrsGlobBrHist = TRUE;
                break;

            case BRPRED_OPT_SINGLE_CPU:
				gBrpredArgs.singleCpuMode = TRUE;
				/* bind all lwp's to single processor */
				bind_lwps();
                break;

            case BRPRED_OPT_U:
                gBrpredArgs.condBrOnly = TRUE;
                break;

            default:
                shade_usage();
                exit(1);
        }
    }
                
            
//		if ( !strncmp( argv[ narg ], "-g", 2 ) ) {
//      
//			char *argString = &( argv[ narg ][ 2 ] );
//			gn_read = sscanf( argString, "%d,%d", &g_value, 
//					  &n_value );	   
//		} else 
//			if ( !strcmp( argv[ narg ], "-u" ) ) {
//				gBrpredArgs.condBrOnly = TRUE;
//			} else	
//				if (!strcmp(argv[narg],"-single-cpu")) {
//					gBrpredArgs.singleCpuMode = TRUE;
//					/* bind all lwp's to single processor */
//					bind_lwps();
//				} else 
//					if (!strcmp(argv[narg],"-pcs")) {
//						gBrpredArgs.perCpuStatistics = TRUE;
//					} else 
//						if ( !strcmp( argv[ narg ], "-r" ) ) {
//							gBrpredArgs.rvrsGlobBrHist = TRUE;
//							
//						} else {
//							shade_usage();
//							exit(1);
//						}
//	}
  
	/* Make sure a new predictor was specified.
	 */
	if (gn_read != 2 || g_value <=0 || n_value <= 0) {
		shade_usage();
		exit(1);
	} else {
		gBrpredArgs.g = g_value;
		gBrpredArgs.n = n_value;
	}


	/* Initialize the branch history compression scheme's various
	 * structures and values (assuming initial state has not yet
	 * been read in).
	 */
	if ( !gBrpredArgs.stateRead ) { 
		for (i=0; i<MAXTHDS; i++) {
			gBrpredVars[i].theBPT =
				(int *) calloc( sizeof( int ), POWER_OF_2( g_value ) );
			if (!gBrpredVars[i].theBPT) {
				shade_error("%s : unable to allocate Branch Prediction Engine buffers\n"
					    "Parameters: g=%d, n=%d\n", 
					    shade_analversion,g_value, n_value );
				exit(1);
			}
		}
	}

	for (i=0; i<MAXTHDS; i++)
		reset_values(gBrpredVars+i, g_value,  n_value);

	(void) shade_trsize(sizeof(shade_trace_t));
	piset = shade_iset_newclass(SHADE_ICLASS_BRANCH, -1);

	/* if single cpu mode, do not trace thread id */
	if (gBrpredArgs.singleCpuMode)
		ptset = shade_tset_new(SHADE_TRCTL_IH,	SHADE_TRCTL_PC, 
				       SHADE_TRCTL_TAKEN, SHADE_TRCTL_EA, -1);
	else
		ptset = shade_tset_new(SHADE_TRCTL_IH,	SHADE_TRCTL_PC, SHADE_TRCTL_TID, 
				       SHADE_TRCTL_TAKEN, SHADE_TRCTL_EA, -1);
	shade_trctl(piset, SHADE_TRI_ISSUED, ptset);
	return 0;
}

pid_t		     pid_parent;

int
shadeuser_analyze() {
	int		   i;
	int		   nins;
	int		   val;
	shade_status_t	   status;
	pid_t		     pid;

	status = shade_appstatus(&val);
	
	switch (status) {
	case SHADE_STATUS_LOADED:
		/*
		 * Remember the parent process ID so we can tell if it forks.
		 */
		pid_parent = getpid();
		break;
	case SHADE_STATUS_FORKED:		
		/* If the application forks, zero the counts in the child.
		 * The child shouldn't inherit execution counts from the
		 * parent.
		 */
		pid = getpid();
		if (pid != pid_parent) {
			pid_parent = pid;
			for (i=0; i<MAXTHDS; i++)
				reset_values(gBrpredVars+i, 
					     gBrpredArgs.g,  
					     gBrpredArgs.n);
				/*
				 * It's no longer important if we lost data due to thread overflow.
				 */
			isThdOvfl = FALSE;
		}
		break;
	default:
		break;
	}
	
	nins = shade_run(gAligner.tb, 10000);
	/* perform data prediction simulation */
	do_bp( gAligner.tb, nins );
	
	return 0;
}

static void
reset_values(BrpredVars* vars, int g, int n) {
	int i;
	vars->theGlobBrHist = 0;
	vars->numMispreds = 0;
	vars->numPredictions = 0;
	/* Initialize the BPT values
	 */
	for ( i = 0; i < POWER_OF_2(g); i++ ) { 
		vars->theBPT[ i ] = 0; /* gBrpredVars.bptTakenValue; */
	}
	vars->brAddrMask = POWER_OF_2(g) - 1;
	vars->globHistMask = POWER_OF_2(g) - 1;
	vars->bptMinValue = 0;
	vars->bptTakenValue = POWER_OF_2(n - 1);
	vars->bptMaxValue = POWER_OF_2(n) - 1;
}

static void
do_bp(shade_trace_t *tb , int ntb) {
  
	int	   i;
	int	   taken; 
	int	   prediction;
	unsigned   tid;

	spix_uint32_t  pc; 
	spix_uint32_t  addrIndex; 
	spix_uint32_t  bptIndex;
  

	for ( i = 0; i < ntb; i++ ) {	 

		/* find out which thread we are in */
		if (gBrpredArgs.singleCpuMode) {
			tid = 0;
		} else {
			tid = tb[i].tr_tid;
			if (tid >= MAXTHDS) {
				isThdOvfl = TRUE;
				continue;
			} else
				if (tid >= nCpusUsed)
					nCpusUsed = tid+1;
		}
		
		pc = tb[ i ].tr_pc;
		taken = ( tb[ i ].tr_taken == 0 ) ? 0 : 1;
    
		/* Is this a branch?
		 */
		if ( spix_sparc_iop_isbranch( tb[ i ].tr_ih ) ) {
      
			/* If this is a conditional branch, predict its outcome
			 * and update the statistics
			 */
			if ( !spix_sparc_iop_isubranch( tb[ i ].tr_ih ) ) {
	
				/* Predict the outcome of this branch for the brpred
				 * branch prediction scheme.  First compute the BPT
				 * index.
				 */
				addrIndex = ( pc >> 2 ) & gBrpredVars[tid].brAddrMask;
				bptIndex = ( gBrpredVars[tid].theGlobBrHist ^ addrIndex );
      
				/* Reference the BPT, make the prediction, and
				 * update statistics.
				 */
				gBrpredVars[tid].numPredictions++;
				prediction = ( gBrpredVars[tid].theBPT[ bptIndex ]
					       >= gBrpredVars[tid].bptTakenValue );
				if ( prediction != taken )
					gBrpredVars[tid].numMispreds++;
	
				/* Update the BPT entry
				 */
				if ( taken ) {
					if ( gBrpredVars[tid].theBPT[ bptIndex ] != 
					     gBrpredVars[tid].bptMaxValue) {
						gBrpredVars[tid].theBPT[ bptIndex ]++;
					}
				} else {
					if ( gBrpredVars[tid].theBPT[ bptIndex ] !=
					     gBrpredVars[tid].bptMinValue ) {
						gBrpredVars[tid].theBPT[ bptIndex ]--;
					}
				}
			}
      
			/* Update the global branch history.  Depending on the command-line
			 * switch, we may update the global branch history for conditional
			 * branches only.
			 */
			if ( spix_sparc_iop_isbranch( tb[ i ].tr_ih ) &&
			     ( !gBrpredArgs.condBrOnly || 
			       !spix_sparc_iop_isubranch( tb[ i ].tr_ih ) ) ) {
				if ( gBrpredArgs.rvrsGlobBrHist ) {
					
					gBrpredVars[tid].theGlobBrHist = 
						( ( gBrpredVars[tid].theGlobBrHist >> 1 ) |
						  ( taken << ( gBrpredArgs.g - 1 ) ) ) &
						gBrpredVars[tid].globHistMask;
					
				} else	
				{
					gBrpredVars[tid].theGlobBrHist = ( ( gBrpredVars[tid].theGlobBrHist 
									     << 1 ) | taken ) 
						& gBrpredVars[tid].globHistMask;
				}
			}
		}
	}
}


void
shadeuser_report(int reason, void* data) 
{
	int	   bptSize;
	char	   bptSizeModifier = ' ';
	unsigned   tid;
	pid_t	   pid;
	int	   g_value = gBrpredArgs.g;
	int	   n_value = gBrpredArgs.n;  
	int	   num_cpus;
	BrpredVars buf_vars, *buf_ptr;
	

	switch (reason) {
	case SHADEUSER_REPORT_TERMINATE:
		/* no info consolidated */
		break;
	case SHADEUSER_REPORT_APP_DONE:
	case SHADEUSER_REPORT_SIGNAL:
		/* Determine the size (in bytes) of the branch prediciton table
		 */
		bptSize = ( ( POWER_OF_2( g_value ) * n_value ) / 8 );
		if ( bptSize > 1024 ) {	 
			bptSize /= 1024;
			bptSizeModifier = 'K';
		}
  
		/*
		 * Warn the user if data was lost due to thread overflow.
		 */
		if (isThdOvfl) {
			shade_error("Warning: Application has too many "
				    "threads, can only analyze first %d threads.\n\n",
				    MAXTHDS);
		}
		fprintf( shade_fp, "\n%s\nBranch Prediction Scheme (%d, %d) Performance Statistics",
			 shade_analversion, g_value, n_value );
		/*
		 * If this is a child process, indicate that in the banner.
		 */
		pid = getpid();
		if (pid != pidFirst)
			fprintf(shade_fp, " [Child Process %d]:\n", pid);
		else
			fprintf(shade_fp, ":\n");

		fprintf( shade_fp, "\n	%d%cByte Branch Prediction Table, %d-bit counters,"
			 " %d-bit global branch history,\n",
			 bptSize, bptSizeModifier, n_value, g_value );
		fprintf( shade_fp, "  %snconditional branches in global branch history\n",
			 gBrpredArgs.condBrOnly ? "No u" : "U" );

		if (gBrpredArgs.perCpuStatistics) {
			num_cpus = nCpusUsed;
			buf_ptr = &gBrpredVars[0];
		} else {
			num_cpus = 1;
			buf_ptr =  &buf_vars;
			buf_vars.numMispreds = buf_vars.numPredictions = 0;
			/* calculate the summary values */
			for (tid = 0; tid < nCpusUsed; tid++) {
				buf_vars.numMispreds += gBrpredVars[tid].numMispreds;
				buf_vars.numPredictions += gBrpredVars[tid].numPredictions;
			}
		}

		for (tid = 0 ; tid < num_cpus; tid++) {
			if (gBrpredArgs.perCpuStatistics && num_cpus > 1 )
				fprintf (shade_fp, "CPU#%d:\n", tid);

			fprintf( shade_fp,
				 "\n%lld mispredicted conditional branches, %lld predictions,"
				 " %.2f%% misprediction ratio\n",
				 buf_ptr[tid].numMispreds, buf_ptr[tid].numPredictions,
				 ( (double) buf_ptr[tid].numMispreds / 
				   (double) buf_ptr[tid].numPredictions ) * 100.0 );
		}

		break;

	default:
		break;
	}
}

int shadeuser_terminate(
	int	ret)
{
	int	val;

	/*
	 * Exit with the termination status from the executed application.
	 */
	switch (shade_appstatus(&val))
	{
	case SHADE_STATUS_EXITED:
		return(val);

	case SHADE_STATUS_SIGNALED:
		return(128 + val);

	case SHADE_STATUS_NOAPP:
		return(0);

	case SHADE_STATUS_LOADED:
	case SHADE_STATUS_RUNNING:
	default:
		assert(0);
	}

}


/*
**  Print a usage message for the command line options that are specific
**  to the 'brpred' analyzer.
*/
void shadeuser_analusage(void)
{
    /*
	fprintf(stderr, "[-g]<g>,<n>	      numbers for the "
		"(g, n) brpred branch prediction scheme\n");
	fprintf(stderr, "-r :\tstore youngest branch outcomes in the MSB of the"
		"global branch history;\n");
	fprintf(stderr, "-single-cpu :\tsimulate all threads on a single cpu;\n");
    */
    shade_print_opt_info(shadeuser_opts, 0);
}

/* binds all lwp's to single processor.
   Necessary for running brpred in single-cpu mode
   on a multiprocessor */
static void		
bind_lwps()
{
	processorid_t i;
	int status;
	int n = sysconf(_SC_NPROCESSORS_ONLN);
	if (n==1) /* nothing to do */
		return;
	for (i=0; ; i++) {
		status = p_online(i, P_STATUS);
		if ((status==-1)&&(errno==EINVAL))
			continue;
		else
			break;
	}
	/* i is valid processor id, bind the process lpws to it */
	status = processor_bind(P_PID, P_MYID, i, NULL);
	if (status == -1)
		shade_error("Was not able to bind Shade process to processor %d, errno == %d\n",
			    i, errno);	
}
