/* Copyright 2006 Sun Microsystems, Inc.  All rights reserved. */

/*
**  icount.c	- Shade analyzer to count dynamic instructions
**
**	This file defines a Shade analyzer that counts the number of
**	dynamic instructions executed by the target application.  The
**	analyzer may optionally count annulled instructions as well as
**	executed instructions.
*/

#pragma ident "@(#)Shade and Spixtools for Solaris 10 v1.0.1"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <shade.h>
#include <shade_anal.h>
#include "getopt.h"

extern char * sim_app_name ();

/*
 * Define a standard trace record.
 */
struct shade_trace_s {SHADE_TRACE};

const char	shadeuser_analversion[] = "icount " "v1.0.1";


/*
 * If -perthread is specified, this is an upper limit on the number of
 * threads we can count separately.  Instruction counts for additional
 * threads are combined.
 */
#define NPERTHD 	50

/*
 * Options specified via command line.
 */
static int	bTraceAnnulled = 0;     	/* trace annulled instructions too */
static int	bPerThread     = 0;     	/* count instructions for each thread */

static int      bSafeExecuted  = 0;             /* TRI EXECUTED via TRI_ISSUED        */

enum {
       ICOUNT_OPT_ANNUL,
       ICOUNT_OPT_PERTHREAD,
       ICOUNT_OPT_SAFE
};

const shade_options_t shadeuser_opts[]={
           {"-annul", "Causes 'icount' to count annulled instructions as well as executed instructions", NULL, NO_PARAM},
           {"-perthread", "Causes 'icount' to display the instruction count for each LWP(thread) in the application separately, rather than combining the counts from all threads together ", NULL, NO_PARAM},
           {"-safe", "Count annulled executed instructions via ANNUL attribute", NULL, NO_PARAM},
           {0, "", NULL, NO_PARAM}
};

static void	init_allthreads(void);
static void	run_allthreads(void);
static void	init_perthread(void);
static void	run_perthread(void);


static pid_t           	        pid_first=-1;
static pid_t           	        pid_parent;

static	unsigned long long	ninsts_thd[NPERTHD];
static	unsigned long long	ninsts_rest;
static	unsigned long long	ninsts_total   = 0;
static	unsigned long long	ninsts_annuled = 0;

static  unsigned long long      ninsts_total_shell = 0;
static  spix_bool_t             shell_mode        = 0;


/*
**  Perform initializations specific to the 'icount' analyzer.  Return zero
**  on success, non-zero on failure.
*/
int shadeuser_initialize(
	int     	argc,
	char ** 	argv,
	char ** 	envp)
{
	spix_bool_t	ok;
	int     	i, j;
    int                icount_optnum;
    shade_option_val_t icount_opt_value;
    

	ok = TRUE;
	for (i = j = 1;  i < argc;  i++) {

        icount_optnum = shade_getoptv(argc, argv, &i, &j, shadeuser_opts, &icount_opt_value);

        switch(icount_optnum){        
            case ICOUNT_OPT_ANNUL:
                bTraceAnnulled = 1;
                break;
             
            case ICOUNT_OPT_PERTHREAD:
                bPerThread = 1;
                break;      
                
            case ICOUNT_OPT_SAFE:
                bSafeExecuted = 1;
                break;
                
            default:
                shade_error("Unknown option '%s'.", argv[i]);
                ok = FALSE;
        }
     }         


/*	ok = TRUE;
	for (i = 1;  i < argc;  i++) {
		if (strcmp(argv[i], "-annul") == 0) {
*/			/*
			 * Count annulled instructions too.
			 */
/*			bTraceAnnulled = 1;
		}
		else if (strcmp(argv[i], "-safe") == 0) {
*/			/*
			 * Count annulled executed instructions via ANNUL attribute.
			 */
/*			bSafeExecuted = 1;
		}

		else if (strcmp(argv[i], "-perthread") == 0) {
*/			/*
			 * Keep separate counts for each thread.
			 */
/*			bPerThread = 1;
		}
		else {
			shade_error("Unknown option '%s'.", argv[i]);
			ok = FALSE;
		}
	}
*/

	if (!ok) {
		shade_usage();
		return(1);
	}

	/*
	 * We want to be notified of fork() operations, so we know if the
	 * application spawns a child process.
	 */
	shade_setopt(SHADE_OPT_FORKNOTIFY);

	/*
	 * Set up the trace control parameters used for all applications
	 * traced by this analyzer.
	 */
	if (bPerThread)
		init_perthread();
	else
		init_allthreads();

	return(0);
}


/*
**  Run and trace an application that is loaded into Shade.
*/
int shadeuser_analyze(void)
{
	if (bPerThread)
		run_perthread();
	else
		run_allthreads();

	return(0);
}


/*
**  Perform termination activities after all applications have been executed.
*/

int shadeuser_terminate(
	int	ret)
{
	int	val;

	/*
	 * Exit with the termination status from the executed application.
	 */
	switch (shade_appstatus(&val))
	{
	case SHADE_STATUS_EXITED:
		return(val);

	case SHADE_STATUS_SIGNALED:
		return(128 + val);

	case SHADE_STATUS_NOAPP:
		return(0);

	case SHADE_STATUS_LOADED:
	case SHADE_STATUS_RUNNING:
	default:
		assert(0);
	}

}



/*
**  Set up the trace control parameters when the "-perthread" options is
**  not specified.  We count all instructions an application executes,
**  combining the counts from all threads.
*/
static void init_allthreads(void)
{
	shade_iset_t *	piset;
	shade_tset_t *	ptset = NULL;

	/*
	 * The size of the trace buffer is not really relevant here since
	 * we aren't writing anything into the trace records.  However, the
	 * size must be more than zero and 8-byte aligned.
	 */

        if (!bSafeExecuted) {
	  shade_trsize(sizeof(shade_trace_t)); 

	  /*
	   * Specify which instructions to trace (all of them).
	   */
	  piset = shade_iset_newclass(SHADE_ICLASS_ANY, -1);

	  /*
	   * If -annul is given, trace executed and annulled instructions.
	   * Otherwise, just trace executed instructions.
	   */

	  if (bTraceAnnulled)
	    shade_trctl(piset, SHADE_TRI_ISSUED, 0); 
	  else
	    shade_trctl(piset, SHADE_TRI_EXECUTED, 0); 
        }
        else {

	  shade_trsize(sizeof(shade_trace_t));
	  piset = shade_iset_newclass(SHADE_ICLASS_ANY, -1);

	  /*
	   * This modification has been done to avoid JVM crash
	   */         

	  if (!bTraceAnnulled)
	    ptset = shade_tset_new(SHADE_TRCTL_ANNULLED, -1);
	  shade_trctl(piset, SHADE_TRI_ISSUED, ptset);

        }


}

/*
**  Run and trace an application when the "-perthread" options is not
**  specified.  The instructions executed by each thread are combined into a
**  single count.
*/
static void run_allthreads(void)
{


	shade_trace_t   	atr[8*1024];
	shade_status_t  	status;
	pid_t           	pid;
	int             	val, i;
        int                     ntr;


	status = shade_appstatus(&val);

	switch (status) {
	case SHADE_STATUS_LOADED:
		/*
		 * Get our process ID, so we can distinguish parent and child in
		 * case the application forks.
		 */
		pid_parent = getpid();
		if (pid_first == -1) {
		  pid_first = pid_parent;
                }

                if (ninsts_total > 0)
                  shell_mode = 1;
  		
		/*
		 * NOTE: We purposely pass an invalid address as the trace buffer
		 * pointer.  Since this analyzer doesn't gather any trace data, the
		 * trace buffer is never written.  Thus, we can pass any properly
		 * aligned address, even an invalid one.  This allows us to pass a
		 * very large buffer without allocating any real memory for it.
		 */

#if 0
		ptr = (shade_trace_t *)8;
	
#endif
		/*
		 * Count instructions until the application terminates.  Continue
		 * counting instructions even if the application loads a new image,
		 * which can happen if the user specified "-exec" on the command
		 * line.
		 */

		ninsts_total = 0;

		break;
	case SHADE_STATUS_FORKED:
		pid = getpid();
		if (pid != pid_parent) {
			pid_parent = pid;
			ninsts_total = 0;
		}

		break;
	}
#if 0
	ninsts_total += shade_run(ptr, 4*1024*1024);
        fprintf (shade_fp, "+");
#endif

        if (bSafeExecuted && !bTraceAnnulled) {
	  ntr = shade_run(atr, sizeof(atr)/sizeof(atr[0]));
	  ninsts_total += ntr;
          ninsts_annuled = 0;
          for (i = 0; i < ntr; i++)
            ninsts_annuled += atr[i].tr_annulled;
          ninsts_total -= ninsts_annuled;  
	}
        else {
	  ninsts_total += shade_run(atr, sizeof(atr)/sizeof(atr[0]));
	}

#if 0
	ninsts_total += shade_run(ptr, 1024*1024);
#endif
}


/*
**  Set up the trace control parameters when the "-perthread" options is
**  specified.  We count all instructions an application executes, keeping
**  separate counts for each thread.
*/
static void init_perthread(void)
{
	shade_iset_t *	piset;
	shade_tset_t *	ptset;


	/*
	 * Set up the size of a trace record.
	 */
	shade_trsize(sizeof(shade_trace_t)); 

	/*
	 * Specify which instructions to trace (all of them).
	 */
	piset = shade_iset_newclass(SHADE_ICLASS_ANY, -1);

	/*
	 * Specify what to trace (just the thread ID).
	 */
	ptset = shade_tset_new(SHADE_TRCTL_TID, -1);

	/*
	 * If -annul is given, trace executed and annulled instructions.
	 * Otherwise, just trace executed instructions.
	 */

#if 0
	if (bTraceAnnulled)
		shade_trctl(piset, SHADE_TRI_ISSUED, 0); 
	else
		shade_trctl(piset, SHADE_TRI_EXECUTED, 0); 
#else

	if (bTraceAnnulled)
		shade_trctl(piset, SHADE_TRI_ISSUED, ptset); 
	else
		shade_trctl(piset, SHADE_TRI_EXECUTED, ptset); 
#endif

}

/*
**  Run and trace an application when the "-perthread" options is specified
**  The instructions executed by each thread are counted separately.
*/
static void run_perthread(void)
{
	shade_trace_t   	atr[8*1024];
	shade_status_t  	status;
	pid_t           	pid;
	int             	val;
	unsigned        	tid;
	unsigned        	ntr;
	unsigned        	i;

	status = shade_appstatus(&val);

	switch (status) {
	case SHADE_STATUS_LOADED:

                if (ninsts_total > 0)
                  shell_mode = 1;


		/*
		 * Get our process ID, so we can distinguish parent and child in
		 * case the application forks.
		 */
		pid_parent = getpid();
		if (pid_first == -1)
			pid_first = pid_parent;
		
		/*
		 * Initialize the instruction counters.
		 */
		memset(ninsts_thd, 0, sizeof(ninsts_thd));
		ninsts_rest = 0;
		ninsts_total = 0;

		break;

	case SHADE_STATUS_FORKED:
		pid = getpid();
		if (pid != pid_parent) {
			pid_parent = pid;
			memset(ninsts_thd, 0, sizeof(ninsts_thd));
			ninsts_rest = 0;
			ninsts_total = 0;
		}
		break;
	}

	/*
	 * Count instructions after shade_run() returns with trace_buffer
	 */

	ntr = shade_run(atr, sizeof(atr)/sizeof(atr[0]));
	ninsts_total += ntr;
	for (i = 0;  i < ntr;  i++) {
		tid = atr[i].tr_tid;
		if (tid < NPERTHD)
			ninsts_thd[tid]++;
		else
			ninsts_rest++;
	}
}


void
shadeuser_report(int reason, void* ptr) {
	pid_t     	pid;
        char            *app_name;  

	switch (reason) {
	case SHADEUSER_REPORT_TERMINATE:
		/* no info is consolidated for applications */
                
                if (shell_mode)
		  fprintf (shade_fp, "Instructions: %llu\n", ninsts_total_shell);

		break;
	  
	case SHADEUSER_REPORT_SIGNAL:		
	case SHADEUSER_REPORT_APP_DONE:

                ninsts_total_shell += ninsts_total;
		
		/*
		 * Print each thread's instruction count.
		 */
		pid = getpid();

                app_name = sim_app_name ();
#if 0
		if (pid != pid_first) {
			fprintf(shade_fp, "Application instructions (child process : %d): "
				"%llu\n", pid, ninsts_total);
#endif

#if 0

		}
		else {
#endif
			fprintf(shade_fp, "Application instructions (%s  pid : %d): %llu\n", app_name, pid, ninsts_total);
#if 0
		}
#endif
		
		if (bPerThread) {
			int i;
			for (i = 0;  i < NPERTHD;  i++) {
				if (ninsts_thd[i]) {
					fprintf(shade_fp, "\tThread %3u icount: %llu\n", i,
						ninsts_thd[i]);
				}
			}
			if (ninsts_rest)
				fprintf(shade_fp, "\tRemaining threads: %llu\n", ninsts_rest);
		}
		break;

	default:
		break;

		
	}
	
	fflush(shade_fp);

}

/*
**  Print a usage message for the command line options that are specific
**  to the 'icount' analyzer.
*/
void shadeuser_analusage(void)
{
    /*
	fprintf(stderr, "%-20s %s\n", "-annul",
	    "Count annulled instructions too");
	fprintf(stderr, "%-20s %s\n", "-perthread",
	    "Keep separate counts for each thread");*/
    shade_print_opt_info(shadeuser_opts, 0);
}









