/* Copyright 2006 Sun Microsystems, Inc.  All rights reserved. */

/*
**  shade.h	- Public interface for Shade analyzers
**
**	This header is the public interface exported to Shade analyzers.
**	Shade is a framework for building tools that trace and analyze
**	executable programs.  Trace information can be collected and
**	dynamically analyzed as the program runs.
**
**	This header defines the processor independent interfaces to Shade.
**	Analyzers that are not dependent on the target processor can
**	include just this header.  However, such analyzers must define the
**	SPIX_ADDR macro to define the size (in bits) of a target address.
**	Analyzers that require processor specific information must include
**	the appropriate processor specific Shade header.  Since the
**	processor specific header defines the target address size, such
**	analyzers need not define SPIX_ADDR.
*/

#ifndef SHADE_H
#define SHADE_H

#pragma ident "@(#)Shade and Spixtools for Solaris 10 v1.0.1"

#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/types.h>
#include <spixtypes.h>

#ifdef __cplusplus
extern "C" {
#endif


/*
 * Possible trace control parameters.  These are the instruction fields
 * that an analyzer can trace.
 */
typedef enum {
	SHADE_TRCTL_PC			= 0,
	SHADE_TRCTL_EA			= 1,
	SHADE_TRCTL_IH			= 2,
	SHADE_TRCTL_TID			= 3,
	SHADE_TRCTL_TAKEN		= 4,
	SHADE_TRCTL_ANNULLED		= 5,
	SHADE_TRCTL_IWSTART		= 6,
	SHADE_TRCTL_STOPB		= 7,
	SHADE_TRCTL_STOPA		= 8
} shade_trctl_t;


/*
 * Instruction classifications.	 These are the classes of instructions
 * that an analyzer can trace.
 */
typedef enum {
	SHADE_ICLASS_ANY		= 0,
	SHADE_ICLASS_IWSTART		= 1,
	SHADE_ICLASS_FP			= 2,
	SHADE_ICLASS_LOAD		= 3,
	SHADE_ICLASS_USTORE		= 4,
	SHADE_ICLASS_CSTORE		= 5,
	SHADE_ICLASS_BRANCH		= 6,
	SHADE_ICLASS_UBRANCH		= 7,
	SHADE_ICLASS_CBRANCH		= 8,
	SHADE_ICLASS_TRAP		= 9
} shade_iclass_t;


/*
 * These are the instruction types that an analyzer can trace.
 */
typedef enum {
	SHADE_TRI_ANNULLED	= (1<<0),
	SHADE_TRI_EXECUTED	= (1<<1),
	SHADE_TRI_ISSUED	= (SHADE_TRI_ANNULLED | SHADE_TRI_EXECUTED)
} shade_tri_t;


/*
 * Possible status values for the traced application.
 */
typedef enum {
	SHADE_STATUS_NOAPP,
	SHADE_STATUS_LOADED,
	SHADE_STATUS_RUNNING,
	SHADE_STATUS_FORKED,
	SHADE_STATUS_EXITED,
	SHADE_STATUS_SIGNALED
} shade_status_t;


/*
 * Possible options that affect the way the application is traced.
 */
typedef enum {
	SHADE_OPT_EXECTRACE,
	SHADE_OPT_EXECNOTRACE,
	SHADE_OPT_FORKNOTIFY,
	SHADE_OPT_NOFORKNOTIFY
} shade_opt_t;


/*
 * SL = 05/23/2001 - current position of TAKEN bit (10 bit) 
 * is a cause of trace error (IH field) when BPR instruction
 * is translated into MOVR
 */

#if 0

/*
 * The definition of the initial fields in any trace record.
 */
# define SHADE_TRACE				            \
	unsigned		tr_ih	            : 16;   \
	unsigned		tr_shade_reserved0  : 5;    \
	unsigned		tr_annulled         : 1;    \
	unsigned		tr_taken            : 1;    \
	unsigned		tr_iwstart          : 1;    \
	unsigned		tr_shade_reserved1  : 8;    \
	unsigned		tr_tid;	                    \
	spix_addr_t		tr_pc;		            \
	spix_addr_t		tr_ea;

#else

/*
 * The definition of the initial fields in any trace record.
 */
# define SHADE_TRACE				            \
	unsigned		tr_ih	            : 16;   \
	unsigned		tr_shade_reserved0  : 13;    \
	unsigned		tr_annulled         : 1;    \
	unsigned		tr_taken            : 1;    \
	unsigned		tr_iwstart          : 1;    \
	unsigned		tr_tid;	                    \
	spix_addr_t		tr_pc;		            \
	spix_addr_t		tr_ea;



#endif

/*
 * Forward declaration of the trace record.  Each analyzer must complete
 * this definition.
 */
typedef struct shade_trace_s	shade_trace_t;


/*
 * Forward declaration of the target processor's register set.	Analyzers
 * should include the processor-specific Shade header to complete this
 * definition.
 */
typedef struct shade_regs_s	shade_regs_t;


/*
 * These types are opaque to all analyzers.
 */
typedef struct shade_iset_s	shade_iset_t;
typedef struct shade_tset_s	shade_tset_t;
typedef struct shade_lock_s	shade_lock_t;
typedef struct shade_rwlock_s	shade_rwlock_t;


/*
 * A version string identifying this version of the Shade kit.
 */
extern const char	shade_version[];


/*
 * Interfaces to manage a set of instruction opcodes.
 */
shade_iset_t *	shade_iset_newclass(shade_iclass_t, ...);
shade_iset_t *	shade_iset_newcopy(shade_iset_t *);
void		shade_iset_free(shade_iset_t *);
shade_iset_t *	shade_iset_addclass(shade_iset_t *, shade_iclass_t, ...);


/*
 * Interfaces to manage a set of trace parameters.
 */
shade_tset_t *	shade_tset_new(shade_trctl_t, ...);
shade_tset_t *	shade_tset_newcopy(shade_tset_t *);
void		shade_tset_free(shade_tset_t *);
shade_tset_t *	shade_tset_add(shade_tset_t *, shade_trctl_t, ...);


/*
 * Interfaces to restrict tracing to address ranges.
 */
void		shade_addtrange(spix_addr_t, spix_addr_t);
void		shade_subtrange(spix_addr_t, spix_addr_t);
spix_bool_t	shade_intrange(spix_addr_t);
int		shade_argtrange(const char *);


/*
 * Interfaces to define trace control.
 */
int	shade_trsize(size_t);
void	shade_trctl(shade_iset_t *, shade_tri_t, shade_tset_t *);
void	shade_trctl_at(spix_addr_t, shade_iset_t *, shade_tri_t,
	    shade_tset_t *);
void	shade_trfun(shade_iset_t *, shade_tri_t,
	    void (*pf_prefun)(
		shade_trace_t *,
		const shade_regs_t *),
	    void (*pf_postfun)(
		shade_trace_t *,
		const shade_regs_t *),
	    void (*pf_unprefun)(
		shade_trace_t *),
	    void (*pf_fixprefun)(
		shade_trace_t *,
		const shade_regs_t *));
void	shade_trfun_at(spix_addr_t, shade_iset_t *, shade_tri_t,
	    void (*pf_prefun)(
		shade_trace_t *,
		const shade_regs_t *),
	    void (*pf_postfun)(
		shade_trace_t *,
		const shade_regs_t *),
	    void (*pf_unprefun)(
		shade_trace_t *),
	    void (*pf_fixprefun)(
		shade_trace_t *,
		const shade_regs_t *));
void	shade_trclear(void);


/*
 * Interfaces to gather trace data.
 */
unsigned	shade_run(shade_trace_t *, unsigned);


/*
 * Interface to parse argument strings.
 */
int	shade_splitargs(char **, char ***, int *);


/*
 * Interfaces to load and query an application.
 */
int		shade_load(const char *, char * const *, char * const *);
int		shade_loadp(const char *, char * const *, char * const *);
void		shade_unload(void);
shade_status_t	shade_appstatus(int *);
char *		shade_bench_memory(void);
const char *	shade_appname(void);
const char *	shade_interpname(void);
spix_addr_t	shade_appbase(void);
spix_addr_t	shade_interpbase(void);

/*
 * Interfaces to implement a shell interface for loading applications.
 */
int	shade_shell(int (*)(int, char **, char **, char **));
int	shade_fshell(FILE *, int (*)(int, char **, char **, char **));
int	shade_sshell(const char *, int (*)(int, char **, char **, char **));


/*
 * Interfaces for manipulating application's I/O.
 */
int	shade_bench_open(const char *, int, mode_t);
int	shade_bench_close(int);
int	shade_bench_dup2(int, int);


/*
 * Interfaces for Unix style signal handling.
 */
int	shade_analsig(int, void (*)(int, siginfo_t *));
int	shade_sendsig(int, const siginfo_t *);


/*
 * Interface for setting some Shade library options.
 */
int	shade_setopt(shade_opt_t);


/*
 * Interfaces to lock critical sections of code in user-trace functions.
 */
shade_lock_t *		shade_lock_new(void);
void			shade_lock_delete(shade_lock_t *);
void			shade_lock_set(shade_lock_t *);
void			shade_lock_clr(shade_lock_t *);
shade_rwlock_t *	shade_rwlock_new(void);
void			shade_rwlock_delete(shade_rwlock_t *);
void			shade_rwlock_rdset(shade_rwlock_t *);
void			shade_rwlock_wrset(shade_rwlock_t *);
void			shade_rwlock_clr(shade_rwlock_t *);


/*
 * Interfaces to allocate memory safely inside user-trace functions.  (May
 * be used elsewhere as well.)
 */
void *	shade_malloc(size_t);
void *	shade_calloc(size_t, size_t);
void *	shade_realloc(void *, size_t);
void	shade_free(void *);


/*
 * The common interface defines this global variable to be the file
 * pointer to which the analyzer should write analysis data.
 * Initialized to &stdout by default.
 */
extern FILE *  	shade_fp;


/*
 *  Get the name of Shade analyzer
 */
const char *	shade_ego(void);

/* Set the name of Shade analyzer, keeps copy of the argument */
void	        set_shade_ego(const char*);

/*
**  Print a usage message to stderr.
*/

void    	shade_usage(void);

/*
**  Print an error message to stderr (not to shade_fp!).
*/

void    	shade_error(const char *, ...);

/*
 * Entry point for Shade analyzers.
 * Shade library provides the default version of this routine. 
 * See the <shade_anal.h> header and shade_anal(3s) manpage for details.
 */
int	shadeuser_main(int, char **, char **);

/*
 *  Old name for the entry point routine.
 *  Provided for compatibility only - do not use it in new programs !
 */
#define shade_main        shadeuser_main

#ifdef __cplusplus
}
#endif

#endif /*SHADE_H*/
